from vagabond.systems import node
from graphviz import Digraph

# This is a class that will be have all the different path finding algorithms
class env:
    
    """
    A class representing an environment.
    Attributes:
    - get_neighbors (function): A function that returns the neighbors of a node.
    - path (list): A list representing the path in the environment.
    Methods:
    - __init__(get_neighbors_func): Initializes the environment with the given get_neighbors function.
    - __getitem__(index): Returns the node at the specified index in the path.
    - __len__(): Returns the length of the path.
    - __repr__(): Returns a string representation of the path.
    - __str__(): Returns a string representation of the path.
    - __iter__(): Returns an iterator for the path.
    - get(): Returns the path as a list of nodes.
    - add(node): Adds a node to the path.
    - remove(node): Removes a node from the path.
    - clear(): Clears the path.
    - get_raw(): Returns an array of values from the path.
    - display_path(filename='graph', directory='.'): Displays the path in a graph and saves it as a PNG file.
    """

    #Functions -- Important Functions
    get_neighbors = None

    #Other Variables
    path = []

    #Constructor
    def __init__(self, get_neighbors_func):
        self.get_neighbors = get_neighbors_func
    
        if not self.get_neighbors:
            raise ValueError("get_neighbors function is not defined")

    # Get a node from the path
    def __getitem__(self, index):
        try:
            return self.path[index]
        except IndexError:
            raise IndexError("Index out of range") from None

    # Get the length of the path
    def __len__(self):
        return len(self.path)
    
    # Print the path
    def __repr__(self):
        return str(self.path)
    
    # Print the path
    def __str__(self):
        return str(self.path)
    
    # Make it iterable
    def __iter__(self):
        return iter(self.path)
    
    # Get the path as a list of nodes
    def get(self):
        return self.path
    
    #Add a node to the path
    def add(self, node):
        if not isinstance(node, node):
            raise TypeError("node must be an instance of the node class. Got '{}' instead".format(type(node)))
        self.path.append(node)

    # Remove a node from the path
    def remove(self, node):
        if not isinstance(node, node):
            raise TypeError("node must be an instance of the node class. Got '{}' instead".format(type(node)))
        self.path.remove(node)

    # Clear the path
    def clear(self):
        self.path = []
    
    # Get array of values from the path
    def get_raw(self):
        if self.path == []:
            return []
        else:
            return [node.value for node in self.path]

    # display the path in a graph
    def display_path(self, filename='graph', directory='.'):

        """
        Display the path as a graph.
        Args:
            filename (str, optional): The name of the file to save the graph as. Defaults to 'graph'.
            directory (str, optional): The directory to save the graph in. Defaults to '.'.
        Returns:
            dot: The graph object.
        Raises:
            ValueError: If the path is empty.
        """

        if not self.path:
            raise ValueError("Path is empty. Cannot display graph.")

        dot = Digraph()
        dot.node_attr.update(shape='rectangle')

        for node in self.path:
            make_str = "Node: " + node.name
            if node.value is not None:
                make_str += "\n----------\nValue: \n" + str(node.value)
            dot.node(node.name, label=make_str)

            if node.parent:
                dot.edge(node.parent.name, node.name)

        filepath = f"{directory}/{filename}"
        dot.render(filepath, format='png', view=True)
        print(f"Graph saved as '{filename}.png' in directory '{directory}'")
        return dot

class astar(env):

    """
    A* algorithm implementation for pathfinding in a graph.
    Args:
        get_neighbors_func (function): A function that returns the neighbors of a given node.
    Attributes:
        path (list): The path found by the A* algorithm.
    Methods:
        path(start_node, end_node): Finds the shortest path from start_node to end_node using the A* algorithm.
    """
    
    """
    Initializes the A* algorithm.
    Args:
        get_neighbors_func (function): A function that returns the neighbors of a given node.
    """
    """
    Finds the shortest path from start_node to end_node using the A* algorithm.
    Args:
        start_node (node): The starting node of the path.
        end_node (node): The end node of the path.
    Returns:
        list: The shortest path from start_node to end_node.
    Raises:
        TypeError: If start_node or end_node is not an instance of the node class.
        ValueError: If start_node or end_node is not defined.
    """

    def __init__(self, get_neighbors_func):
        super().__init__(get_neighbors_func)
        
    def path(self, start_node = None, end_node = None):

        """
        Finds the shortest path between two nodes in the graph.
        Parameters:
            start_node (node): The starting node of the path.
            end_node (node): The ending node of the path.
        Returns:
            list: The shortest path as a list of nodes.
        Raises:
            TypeError: If start_node or end_node is not an instance of the node class.
            ValueError: If start_node or end_node is not defined.
        """
        
        if start_node is not None:
            if not isinstance(start_node, node):
                raise TypeError("start_node must be an instance of the node class. Got '{}' instead".format(type(start_node)))
        else:
            raise ValueError("start_node is not defined")
        
        if end_node is not None:
            if not isinstance(end_node, node):
                raise TypeError("end_node must be an instance of the node class. Got '{}' instead".format(type(end_node)))
        else:
            raise ValueError("end_node is not defined")
        
        unvisited = []
        visited = []

        # Add the start node to the unvisited list
        unvisited.append(start_node)

        while unvisited:

            # Get the node with the minimum cost
            current_node = min(unvisited, key=lambda x: x.cost)

            # Remove the current node from the unvisited list
            unvisited.remove(current_node)

            # Add the current node to the visited list
            visited.append(current_node)

            # Check if the current node is the end node
            if current_node == end_node:
                path = []
                while current_node:
                    path.append(current_node)
                    current_node = current_node.parent

                self.path = path[::-1]
                return self.path

            # Get the neighbors of the current node
            neighbors = self.get_neighbors(current_node)

            # Loop through the neighbors
            for neighbor_node in neighbors:
                
                # Check if the neighbor is in the visited list
                if neighbor_node in visited:
                    continue

                # Check if the neighbor is in the unvisited list
                if neighbor_node not in unvisited:
                    unvisited.append(neighbor_node)
                else:
                    # Replace the neighbor node with the new node if the new node has a lower cost
                    for i, node_ in enumerate(unvisited):
                        if node_ == neighbor_node and node_.cost > neighbor_node.cost:
                            unvisited[i] = neighbor_node
        return None

class bfs(env):

    """
    Initializes a breadth-first search algorithm.
    Parameters:
    - get_neighbors_func: A function that takes a node as input and returns a list of its neighboring nodes.
    """
    """
    Finds the shortest path from start_node to end_node using breadth-first search algorithm.
    Parameters:
    - start_node: The starting node of the path. Must be an instance of the node class.
    - end_node: The ending node of the path. Must be an instance of the node class.
    Returns:
    - path: A list of nodes representing the shortest path from start_node to end_node.
        Returns None if no path is found.
    """

    def __init__(self, get_neighbors_func):
        super().__init__(get_neighbors_func)
        
    def path(self, start_node=None, end_node=None):

        """
        Finds the path from start_node to end_node using breadth-first search algorithm.
        Args:
            start_node (node, optional): The starting node. Defaults to None.
            end_node (node, optional): The target node. Defaults to None.
        Raises:
            TypeError: If start_node or end_node is not an instance of the node class.
            ValueError: If start_node or end_node is not defined.
        Returns:
            list: The path from start_node to end_node, represented as a list of nodes.
                Returns None if no path is found.
        """

        if start_node is not None:
            if not isinstance(start_node, node):
                raise TypeError("start_node must be an instance of the node class. Got '{}' instead".format(type(start_node)))
        else:
            raise ValueError("start_node is not defined")
        
        if end_node is not None:
            if not isinstance(end_node, node):
                raise TypeError("end_node must be an instance of the node class. Got '{}' instead".format(type(end_node)))
        else:
            raise ValueError("end_node is not defined")
        
        queue = [start_node]
        visited = set()
        parent_map = {start_node: None}

        while queue:
            current_node = queue.pop(0)
            if current_node == end_node:
                path = []
                while current_node:
                    path.append(current_node)
                    current_node = parent_map[current_node]
                self.path = path[::-1]
                return self.path
            
            visited.add(current_node)
            neighbors = self.get_neighbors(current_node)
            for neighbor in neighbors:
                if neighbor not in visited and neighbor not in queue:
                    parent_map[neighbor] = current_node
                    queue.append(neighbor)
        return None

class dfs(env):

    """
    Initializes a depth-first search algorithm.
    Parameters:
    - get_neighbors_func: A function that returns the neighbors of a given node.
    Returns:
    None
    """
    """
    Finds a path from the start node to the end node using depth-first search.
    Parameters:
    - start_node: The starting node of the path. Must be an instance of the node class.
    - end_node: The ending node of the path. Must be an instance of the node class.
    Returns:
    - path: A list of nodes representing the path from the start node to the end node.
        Returns None if no path is found.
    """

    def __init__(self, get_neighbors_func):
        super().__init__(get_neighbors_func)
        
    def path(self, start_node=None, end_node=None):

        """
        Finds a path from the start_node to the end_node using a depth-first search algorithm.
        Args:
            start_node (node): The starting node of the path.
            end_node (node): The target node to reach.
        Returns:
            list: A list of nodes representing the path from start_node to end_node, 
                or None if no path is found.
        Raises:
            TypeError: If start_node or end_node is not an instance of the node class.
            ValueError: If start_node or end_node is not defined.
        """

        if start_node is not None:
            if not isinstance(start_node, node):
                raise TypeError("start_node must be an instance of the node class. Got '{}' instead".format(type(start_node)))
        else:
            raise ValueError("start_node is not defined")
        
        if end_node is not None:
            if not isinstance(end_node, node):
                raise TypeError("end_node must be an instance of the node class. Got '{}' instead".format(type(end_node)))
        else:
            raise ValueError("end_node is not defined")
        
        stack = [start_node]
        visited = set()
        parent_map = {start_node: None}

        while stack:
            current_node = stack.pop()
            if current_node == end_node:
                path = []
                while current_node:
                    path.append(current_node)
                    current_node = parent_map[current_node]
                self.path = path[::-1]
                return self.path
            
            visited.add(current_node)
            neighbors = self.get_neighbors(current_node)
            for neighbor in neighbors:
                if neighbor not in visited and neighbor not in stack:
                    parent_map[neighbor] = current_node
                    stack.append(neighbor)
        return None

class dijkstra(env):

    """
    Initializes a Dijkstra object.
    Parameters:
    - get_neighbors_func (function): A function that returns the neighbors of a given node.
    Returns:
    - None
    """
    """
    Finds the shortest path between two nodes using Dijkstra's algorithm.
    Parameters:
    - start_node (node): The starting node of the path.
    - end_node (node): The ending node of the path.
    Returns:
    - list: The shortest path from start_node to end_node, represented as a list of nodes.
    - None: If no path is found.
    """

    def __init__(self, get_neighbors_func):
        super().__init__(get_neighbors_func)
        
    def path(self, start_node=None, end_node=None):
        
        """
        Finds the shortest path between two nodes in the graph.
        Args:
            start_node (node): The starting node of the path.
            end_node (node): The ending node of the path.
        Returns:
            list: The shortest path as a list of nodes.
        Raises:
            TypeError: If start_node or end_node is not an instance of the node class.
            ValueError: If start_node or end_node is not defined.
        """
        
        if start_node:
            if not isinstance(start_node, node):
                raise TypeError("start_node must be an instance of the node class. Got '{}' instead".format(type(start_node)))
        else:
            raise ValueError("start_node is not defined")
        
        if end_node:
            if not isinstance(end_node, node):
                raise TypeError("end_node must be an instance of the node class. Got '{}' instead".format(type(end_node)))
        else:
            raise ValueError("end_node is not defined")
        
        unvisited = []
        visited = []
        start_node.cost = 0
        unvisited.append(start_node)

        while unvisited:
            current_node = min(unvisited, key=lambda x: x.cost)
            unvisited.remove(current_node)
            visited.append(current_node)

            if current_node == end_node:
                path = []
                while current_node:
                    path.append(current_node)
                    current_node = current_node.parent
                self.path = path[::-1]
                return self.path

            neighbors = self.get_neighbors(current_node)
            for neighbor in neighbors:
                if neighbor in visited:
                    continue
                new_cost = current_node.cost + neighbor.cost
                if neighbor not in unvisited or new_cost < neighbor.cost:
                    neighbor.cost = new_cost
                    neighbor.parent = current_node
                    if neighbor not in unvisited:
                        unvisited.append(neighbor)
        return None



class node:

    """
    A class representing a node in a tree structure.

    Attributes:
    - value: The value associated with the node.
    - childs: A list of child nodes.
    - parent: The parent node.
    - cost: The cost associated with the node.
    - name: The name of the node.

    Methods:
    - __init__(self, value=None, childs=[], parent=None, cost=0, name=None): Initializes a node object.
    - __repr__(self) -> str: Returns a string representation of the node.
    - __eq__(self, o: object) -> bool: Checks if two nodes are equal.
    - __add__(self, b): Adds a node as a child to the current node.
    - __radd__(self, a): Adds the current node as a child to another node.
    - __sub__(self, b): Removes a child node from the current node.
    - __rsub__(self, a): Removes the current node from the child list of another node.
    - __hash__(self): Makes the node hashable.
    - add_child(self, child): Adds a child node to the current node.
    - remove_child(self, child): Removes a child node from the current node.
    - clear(self): Clears the list of child nodes.
    - get_children(self): Returns the list of child nodes.
    - get_parent(self): Returns the parent node.
    - set_parent(self, parent): Sets the parent node.
    - get_cost(self): Returns the cost of the node.
    - set_cost(self, cost): Sets the cost of the node.
    - get_name(self): Returns the name of the node.
    - set_name(self, name): Sets the name of the node.
    - get_value(self): Returns the value of the node.
    - set_value(self, value): Sets the value of the node.
    """

    childs = []
    parent = None
    cost = 0
    name = ""
    value = None

    #Constructor
    def __init__(self, value=None, childs=[], parent=None, cost=0, name=None):

        """
        Initialize a Vagabond system object.
        Parameters:
        - value (optional): The value of the system.
        - childs (optional): A list of child systems.
        - parent (optional): The parent system.
        - cost (optional): The cost of the system.
        - name (optional): The name of the system.
        """

        self.childs = childs
        self.value = value
        self.name = name
        self.parent = parent
        self.cost = cost

    #print function / string representation
    def __repr__(self) -> str:

        """
        Returns a string representation of the object.
        If the object has a name attribute, it returns the name.
        If the object does not have a name attribute, it tries to convert the value attribute to a string and returns it.
        If the conversion fails or the object does not have a value attribute, it returns "un-named node".
        """

        if(self.name):
            return self.name
        else:
            try:
                return str(self.value)
            except:
                return "un-named node"
        
    #equality function
    def __eq__(self, o: object) -> bool:

        """
        Compare the value of this object with another object.

        Parameters:
            o (object): The object to compare with.

        Returns:
            bool: True if the values are equal, False otherwise.
        """
        
        return self.value == o.value
    
    # Addition operation - add node b as a child to node a
    def __add__(self, b):

        """
        Adds a node to the list of child nodes.

        Parameters:
            b (node): The node to be added.

        Returns:
            self: The updated object with the added node.

        Raises:
            TypeError: If the operand type is not supported.

        """

        if isinstance(b, node):
            self.childs.append(b)
            return self
        else:
            raise TypeError("unsupported operand type(s) for +: '{}' and '{}'".format(type(self), type(b)))
        
    # Reverse addition operation - add node a as a child to node b
    def __radd__(self, a):

        """
        Right addition method for the 'node' class.

        Parameters:
        - a: An object of type 'node' or its subclass.

        Returns:
        - If 'a' is an instance of 'node', appends 'self' to the 'childs' list of 'a' and returns 'a'.
        - Raises a TypeError if 'a' is not an instance of 'node'.

        """

        if isinstance(a, node):
            a.childs.append(self)
            return a
        else:
            raise TypeError("unsupported operand type(s) for +: '{}' and '{}'".format(type(a), type(self)))
        
    # Subtraction operation - remove node b from the child list of node a
    def __sub__(self, b):

        """
        Subtract a node from the current node's list of children.

        Parameters:
        - b: The node to be subtracted from the current node's children.

        Returns:
        - self: The current node after removing the specified node from its children.

        Raises:
        - ValueError: If the specified node is not a child of the current node.
        - TypeError: If the operand type is not supported.

        """

        if isinstance(b, node):
            if b in self.childs:
                self.childs.remove(b)
                return self
            else:
                raise ValueError("node '{}' is not a child of '{}'".format(b.name, self.name))
        else:
            raise TypeError("unsupported operand type(s) for -: '{}' and '{}'".format(type(self), type(b)))
    
    # Subtraction operation - remove node b from the child list of node a
    def __rsub__(self, a):

        """
        Subtract a node from the current node's list of children.

        Parameters:
        - a (node): The node to be subtracted from the current node's children.

        Returns:
        - a (node): The subtracted node.

        Raises:
        - ValueError: If the given node is not a child of the current node.
        - TypeError: If the operand types are not supported for subtraction.
        """

        if isinstance(a, node):
            if a in self.childs:
                self.childs.remove(a)
                return a
            else:
                raise ValueError("node '{}' is not a child of '{}'".format(a.name, self.name))
        else:
            raise TypeError("unsupported operand type(s) for -: '{}' and '{}'".format(type(a), type(self)))

    # Make it hashable
    def __hash__(self):

        """
        Return the hash value of the object.

        Returns:
            int: The hash value of the object.
        """

        return hash(self.value)
    
    # Add a child to the node
    def add_child(self, child):

        """
        Adds a child node to the current node.

        Parameters:
        - child: An instance of the node class representing the child node.

        Raises:
        - TypeError: If the provided child is not an instance of the node class.

        Returns:
        - None
        """

        if not isinstance(child, node):
            raise TypeError("child must be an instance of the node class. Got '{}' instead".format(type(child)))
        self.childs.append(child)

    # Remove a child from the node
    def remove_child(self, child):

        """
        Removes a child node from the current node.

        Parameters:
            child (node): The child node to be removed.

        Raises:
            TypeError: If the `child` parameter is not an instance of the `node` class.

        Returns:
            None
        """

        if not isinstance(child, node):
            raise TypeError("child must be an instance of the node class. Got '{}' instead".format(type(child)))
        self.childs.remove(child)

    # Clear the children list
    def clear(self):

        """
        Clears the list of child elements.

        Parameters:
            None

        Returns:
            None
        """

        self.childs = []

    # Get the children list
    def get_children(self):

        """
        Returns the children of the current object.

        Returns:
            list: A list of child objects.
        """

        return self.childs
    
    # Get the parent node
    def get_parent(self):

        """
        Returns the parent of the current object.

        Returns:
            The parent object.
        """

        return self.parent
    
    # Set the parent node
    def set_parent(self, parent):

        """
        Sets the parent of the current node.

        Parameters:
        - parent: An instance of the node class.

        Raises:
        - TypeError: If the parent is not an instance of the node class.

        Returns:
        - None
        """

        if not isinstance(parent, node):
            raise TypeError("parent must be an instance of the node class. Got '{}' instead".format(type(parent)))
        self.parent = parent

    # Get the cost of the node
    def get_cost(self):

        """
        Returns the cost of the object.

        :return: The cost of the object.
        :rtype: float
        """

        return self.cost
    
    # Set the cost of the node
    def set_cost(self, cost):

        """
        Set the cost of the system.

        Parameters:
        - cost: The cost value to be set.

        Returns:
        None
        """

        self.cost = cost

    # Get the name of the node
    def get_name(self):

        """
        Returns the name of the object.

        Returns:
            str: The name of the object.
        """

        return self.name
    
    # Set the name of the node
    def set_name(self, name):

        """
        Sets the name of the object.

        Parameters:
        - name (str): The name to be set.

        Returns:
        None
        """

        self.name = name

    # Get the value of the node
    def get_value(self):

        """
        Returns the value of the object.

        Returns:
            The value of the object.
        """

        return self.value
    
    # Set the value of the node
    def set_value(self, value):

        """
        Set the value of the object.

        Parameters:
        - value: The value to be set.

        Returns:
        None
        """
        
        self.value = value



from setuptools import setup, find_packages

setup(
    name='py-vagabond',
    version='2.0.2',
    license="MIT License with attribution requirement",
    author="Ranit Bhowmick",
    author_email='bhowmickranitking@duck.com',
    description='''Vagabond is a comprehensive library for pathfinding, navigation, and environmental understanding in robotics and automation. It provides a range of algorithms and tools to help developers create efficient and sophisticated pathfinding solutions for robots and autonomous systems.''',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/Kawai-Senpai/Vagabond',
    download_url='https://github.com/Kawai-Senpai/Vagabond',
    keywords=["Pathfinding", "Robotics", "Navigation", "A* Algorithm", "Graph Theory", "Automation", "Robotic Systems"],
    install_requires=['graphviz==0.20.1'],
)
